#ifndef _DIR_SAMPLED_H_INCLUDED_
#define _DIR_SAMPLED_H_INCLUDED_

#include "matrix.h"
#include "mex.h"
#include <math.h>
#include "array.h"

#include "helperMEX.h"
#include "debugMEX.h"

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include "categorical.h"
#include <omp.h>
#include "tableFuncs.h"

#include <boost/unordered_map.hpp>
using boost::unordered_map;

class dir_sampled
{
   public:
   int N;
   int Nthreads;
   arr(gsl_rng*) rArr;

   // prior hyperparameters
   Real alphah;
   int D;

   // sufficient statistics of the observed data
   // assume sparse
   unordered_map<int, int> indices_counts;

   // instantiated categorical parameters
   categorical param;
   Real logmu_prior, logmu_posterior;


   // temporary space for parallel traversal of unordered_map
   int** temp_counts_itr;
   int* temp_indices_itr;
   int temp_counts_N;
   bool temp_counts_invalid;


   double (*gammaln)(double);



public:
   // --------------------------------------------------------------------------
   // -- dir_sampled
   // --   constructor; initializes to empty
   // --------------------------------------------------------------------------
   dir_sampled();
   // --------------------------------------------------------------------------
   // -- dir_sampled
   // --   copy constructor;
   // --------------------------------------------------------------------------
   dir_sampled(const dir_sampled& that);
   // --------------------------------------------------------------------------
   // -- operator=
   // --   assignment operator
   // --------------------------------------------------------------------------
   dir_sampled& operator=(const dir_sampled& that);
   // --------------------------------------------------------------------------
   // -- copy
   // --   returns a copy of this
   // --------------------------------------------------------------------------
   void copy(const dir_sampled& that, bool isClean=true, bool copyIndices_counts=true);
   // --------------------------------------------------------------------------
   // -- dir_sampled
   // --   constructor; intializes to all the values given
   // --------------------------------------------------------------------------
   dir_sampled(int _Nthreads, int _D, Real _alphah);

   // --------------------------------------------------------------------------
   // -- ~dir_sampled
   // --   destructor
   // --------------------------------------------------------------------------
   virtual ~dir_sampled();

   // --------------------------------------------------------------------------
   // -- ~cleanup
   // --   deletes all the memory allocated by this
   // --------------------------------------------------------------------------
   virtual void cleanup();

public:
   // --------------------------------------------------------------------------
   // -- empty
   // --   Empties out the statistics of the dir_sampled (i.e. no data).
   // --------------------------------------------------------------------------
   void clear();
   void empty();

   bool isempty() const;
   int getN() const;
   int getD() const;
   arr(Real) get_logpi() const;
   gsl_rng* get_r();

   void set_categorical(categorical &other);
   void set_categorical(arr(Real) _logpi);
   categorical* get_categorical();

   // --------------------------------------------------------------------------
   // -- update_posteriors
   // --   Updates the posterior hyperparameters
   // --------------------------------------------------------------------------
   void update_posteriors();
   void update_posteriors_sample();

   // helper function to merge to indices_counts
   int merge_indices_counts(unordered_map<int,int> &indices_counts2, bool clear=true);

   void temp_merge_with(dir_sampled* other);
   void merge_with(dir_sampled* other, bool doSample);
   void merge_with(dir_sampled* other1, dir_sampled* other2, bool doSample);
   void merge_with(dir_sampled other, bool doSample);
   void merge_with(dir_sampled other1, dir_sampled other2, bool doSample);
   void set_stats(int _N, const unordered_map<int,int> &_indices_counts);
   void set_stats(int _N, const unordered_map<int,int> &indices_counts1, const unordered_map<int,int> &indices_counts2);
   void set_stats(int _N, const unordered_map<int,int> &indices_counts1, const unordered_map<int,int> &indices_counts2,
      unordered_map<int,int> &indices_counts3);

#ifdef USELOGLIKELIHOOD
   Real predictive_loglikelihood(arr(long) indices, arr(Real) values, long nnz) const;
   Real predictive_loglikelihood(unordered_map<int,int> &indices_counts2) const;
   Real predictive_loglikelihood(long index) const;
#else
   Real predictive_likelihood(arr(long) indices, arr(Real) values, long nnz) const;
   Real predictive_likelihood(unordered_map<int,int> &indices_counts2) const;
   Real predictive_likelihood(long index) const;
#endif

   Real predictive_loglikelihood_marginalized(arr(long) indices, arr(Real) values, long nnz);
   Real predictive_loglikelihood_marginalized(unordered_map<int,int> &indices_counts2);
   Real predictive_loglikelihood_marginalized(long index);

   void sample();

   Real data_loglikelihood() const;
   Real data_loglikelihood_marginalized();
   Real data_loglikelihood_marginalized_testmerge(dir_sampled *other) const;


   Real getlogmu_posterior();
   Real getlogmu_prior() const;

   Real Jdivergence(const dir_sampled &other);

   // --------------------------------------------------------------------------
   // -- add_data
   // --   functions to add an observation to the dir_sampled. Updates the sufficient
   // -- statistics, posterior hyperparameters, and predictive parameters
   // --
   // --   parameters:
   // --     - data : the new observed data point of size [1 D]
   // --------------------------------------------------------------------------
   void rem_data(arr(long) indices, arr(Real) values, long nnz);
   void add_data_init(arr(long) indices, arr(Real) values, long nnz);
   void add_data(arr(long) indices, arr(Real) values, long nnz);

   void rem_data(unordered_map<int,int> &indices_counts2);
   void add_data(unordered_map<int,int> &indices_counts2);

   void rem_data_dec(long index);
   void add_data_inc_init(long index);
   void add_data_inc(long index);

   void prepare_indices_counts_itr();
};



#ifdef USELOGLIKELIHOOD
inline Real dir_sampled::predictive_loglikelihood(arr(long) indices, arr(Real) values, long nnz) const
{
   return param.predictive_loglikelihood(indices, values, nnz);
}
inline Real dir_sampled::predictive_loglikelihood(unordered_map<int,int> &indices_counts2) const
{
   return param.predictive_loglikelihood(indices_counts2);
}

inline Real dir_sampled::predictive_loglikelihood(long index) const
{
   return param.predictive_loglikelihood(index);
}

#else

inline Real dir_sampled::predictive_likelihood(arr(long) indices, arr(Real) values, long nnz) const
{
   return param.predictive_likelihood(indices, values, nnz);
}
inline Real dir_sampled::predictive_likelihood(unordered_map<int,int> &indices_counts2) const
{
   return param.predictive_likelihood(indices_counts2);
}
inline Real dir_sampled::predictive_likelihood(long index) const
{
   return param.predictive_likelihood(index);
}
#endif




inline Real dir_sampled::predictive_loglikelihood_marginalized(arr(long) indices, arr(Real) values, long nnz)
{
   Real log_likelihood = 0;
   Real A = alphah*D;
   for (unordered_map<int,int>::iterator itr=indices_counts.begin(); itr!=indices_counts.end(); itr++)
   {
      A += itr->second;
   }

   int C = 0;
   for (long di=0; di<nnz; di++)
   {
      long ind = indices[di];
      Real val = values[di];
      C += val;
      Real alpha = alphah;
      unordered_map<int,int>::const_iterator itr = indices_counts.find(ind);
      if (itr != indices_counts.end()) // is it found?
         alpha += itr->second;
      log_likelihood += gammaln(val + alpha) - gammaln(alpha) - gammaln(val+1);
   }
   log_likelihood += gammaln(C+1) + gammaln(A) - gammaln(C+A);
   
   return log_likelihood;
}

inline Real dir_sampled::predictive_loglikelihood_marginalized(unordered_map<int,int> &indices_counts2)
{
   Real sumPostAlpha = alphah*D + N;
   Real loglikelihood = gammaln(sumPostAlpha);

   int totalCount = 0;
   for (unordered_map<int,int>::iterator it = indices_counts2.begin(); it!=indices_counts2.end(); it++)
   {
      int index = it->first;
      int count = it->second;
      totalCount += count;

      Real postAlpha = alphah;
      unordered_map<int,int>:: iterator ito = indices_counts.find(index);
      if (ito!=indices_counts.end())
         postAlpha += ito->second;

      loglikelihood += gammaln(postAlpha + count) - gammaln(postAlpha);
   }
   loglikelihood -= gammaln(totalCount + sumPostAlpha);

   return loglikelihood;
}
inline Real dir_sampled::predictive_loglikelihood_marginalized(long index)
{
   Real sumPostAlpha = alphah*D + N;
   Real loglikelihood = -log(sumPostAlpha-1);

   unordered_map<int,int>::iterator itr = indices_counts.find(index);
   if (itr == indices_counts.end())
      loglikelihood += log(alphah);
   else
      loglikelihood += log(alphah+itr->second);

   return loglikelihood;
}



#endif
