#ifndef _DIR_SAMPLED_FULL_H_INCLUDED_
#define _DIR_SAMPLED_FULL_H_INCLUDED_

#include "matrix.h"
#include "mex.h"
#include <math.h>
#include "array.h"

#include "helperMEX.h"
#include "debugMEX.h"

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include "categorical.h"
#include <omp.h>
#include "tableFuncs.h"

#include <vector>
using std::vector;

class dir_sampled_full
{
   public:
   int N;
   int Nthreads;
   arr(gsl_rng*) rArr;

   // prior hyperparameters
   Real alphah;
   int D;

   // sufficient statistics of the observed data
   // assume sparse
   vector<int> indices_counts;

   // instantiated categorical parameters
   categorical param;
   Real logmu_prior, logmu_posterior;

   double (*gammaln)(double);



public:
   // --------------------------------------------------------------------------
   // -- dir_sampled_full
   // --   constructor; initializes to empty
   // --------------------------------------------------------------------------
   dir_sampled_full();
   // --------------------------------------------------------------------------
   // -- dir_sampled_full
   // --   copy constructor;
   // --------------------------------------------------------------------------
   dir_sampled_full(const dir_sampled_full& that);
   // --------------------------------------------------------------------------
   // -- operator=
   // --   assignment operator
   // --------------------------------------------------------------------------
   dir_sampled_full& operator=(const dir_sampled_full& that);
   // --------------------------------------------------------------------------
   // -- copy
   // --   returns a copy of this
   // --------------------------------------------------------------------------
   void copy(const dir_sampled_full& that, bool isClean=true, bool copyIndices_counts=true);
   void copyNoNewMem(const dir_sampled_full& that, bool copyIndices_counts=true);
   // --------------------------------------------------------------------------
   // -- dir_sampled_full
   // --   constructor; intializes to all the values given
   // --------------------------------------------------------------------------
   dir_sampled_full(int _Nthreads, int _D, Real _alphah);

   // --------------------------------------------------------------------------
   // -- ~dir_sampled_full
   // --   destructor
   // --------------------------------------------------------------------------
   virtual ~dir_sampled_full();

   // --------------------------------------------------------------------------
   // -- ~cleanup
   // --   deletes all the memory allocated by this
   // --------------------------------------------------------------------------
   virtual void cleanup();

public:
   // --------------------------------------------------------------------------
   // -- empty
   // --   Empties out the statistics of the dir_sampled_full (i.e. no data).
   // --------------------------------------------------------------------------
   void clear();
   void empty();

   bool isempty() const;
   int getN() const;
   int getD() const;
   arr(Real) get_logpi() const;
   gsl_rng* get_r();

   void set_categorical(categorical &other);
   void set_categorical(arr(Real) _logpi);
   categorical* get_categorical();

   // --------------------------------------------------------------------------
   // -- update_posteriors
   // --   Updates the posterior hyperparameters
   // --------------------------------------------------------------------------
   void update_posteriors();
   void update_posteriors_sample();

   // helper function to merge to indices_counts
   int merge_indices_counts(vector<int> &indices_counts2, bool clear=true);

   void temp_merge_with(dir_sampled_full* other);
   void merge_with(dir_sampled_full* other, bool doSample);
   void merge_with(dir_sampled_full* other1, dir_sampled_full* other2, bool doSample);
   void merge_with(dir_sampled_full other, bool doSample);
   void merge_with(dir_sampled_full other1, dir_sampled_full other2, bool doSample);
   void set_stats(int _N, const int* _indices_counts);
   void set_stats(int _N, const int* indices_counts1, const int* indices_counts2);
   void set_stats(int _N, const int* indices_counts1, const int* indices_counts2,
      const int* indices_counts3);

#ifdef USELOGLIKELIHOOD
   Real predictive_loglikelihood(arr(long) indices, arr(Real) values, long nnz) const;
   Real predictive_loglikelihood(long index) const;
#else
   Real predictive_likelihood(arr(long) indices, arr(Real) values, long nnz) const;
   Real predictive_likelihood(long index) const;
#endif

   Real predictive_loglikelihood_marginalized(arr(long) indices, arr(Real) values, long nnz);
   Real predictive_loglikelihood_marginalized(vector<int> &indices_counts2);
   Real predictive_loglikelihood_marginalized(long index);

   void sample();

   Real data_loglikelihood() const;
   Real data_loglikelihood_marginalized();
   Real data_loglikelihood_marginalized_testmerge(const dir_sampled_full *other) const;
   Real data_loglikelihood_marginalized_testmerge(const int otherN, const int* otherindices_counts) const;


   Real getlogmu_posterior();
   Real getlogmu_prior() const;

   Real Jdivergence(const dir_sampled_full &other);

   // --------------------------------------------------------------------------
   // -- add_data
   // --   functions to add an observation to the dir_sampled_full. Updates the sufficient
   // -- statistics, posterior hyperparameters, and predictive parameters
   // --
   // --   parameters:
   // --     - data : the new observed data point of size [1 D]
   // --------------------------------------------------------------------------
   void rem_data(vector<int> &indices_counts2);
   void add_data(vector<int> &indices_counts2);
   void rem_data(int* indices_counts2);
   void add_data(int* indices_counts2);

   void rem_data_dec(long index);
   void add_data_inc_init(long index);
   void add_data_inc(long index);
};



#ifdef USELOGLIKELIHOOD
inline Real dir_sampled_full::predictive_loglikelihood(arr(long) indices, arr(Real) values, long nnz) const
{
   return param.predictive_loglikelihood(indices, values, nnz);
}
inline Real dir_sampled_full::predictive_loglikelihood(long index) const
{
   return param.predictive_loglikelihood(index);
}

#else

inline Real dir_sampled_full::predictive_likelihood(arr(long) indices, arr(Real) values, long nnz) const
{
   return param.predictive_likelihood(indices, values, nnz);
}
inline Real dir_sampled_full::predictive_likelihood(vector<int> &indices_counts2) const
{
   return param.predictive_likelihood(indices_counts2);
}
inline Real dir_sampled_full::predictive_likelihood(long index) const
{
   return param.predictive_likelihood(index);
}
#endif


inline Real dir_sampled_full::predictive_loglikelihood_marginalized(vector<int> &indices_counts2)
{
   Real sumPostAlpha = alphah*D + N;
   Real loglikelihood = gammaln(sumPostAlpha);

   int totalCount = 0;
   for (int d=0; d<D; d++)
   {
      int count = indices_counts2[d];
      totalCount += count;

      Real postAlpha = alphah + indices_counts[d];

      loglikelihood += gammaln(postAlpha + count) - gammaln(postAlpha);
   }
   loglikelihood -= gammaln(totalCount + sumPostAlpha);

   return loglikelihood;
}

inline Real dir_sampled_full::predictive_loglikelihood_marginalized(long index)
{
   Real sumPostAlpha = alphah*D + N;
   Real loglikelihood = -log(sumPostAlpha-1);
   loglikelihood += log(alphah + indices_counts[index]);

   return loglikelihood;
}



#endif
