function [model, E, EHOW, time, numTopics, nhists] = run_HDP_subclusters(data, dictSize, start, dispOn, Mproc, gamma, alpha, endtime, numits, isBars)
% RUN_HDPMM_SUBCLUSTERS - runs the HDP with subcluster splits and merges
%    run_HDP_subclusters(data, dictSize, start, dispOn, Mproc, gamma,
%     alpha, endtime, numits, isBars)
%
%    data - a 1xD struct array, where D is the number of documents. Each
%       element in the array should have one field named "words", which
%       contains a 1D array of indices of words in the corresponding
%       document. Words should have indices starting at 0, and have a
%       maximum value of dictSize - 1.
%    dictSize - the size of the dictionary
%    start - the number of initial clusters to start with
%    dispOn - whether progress should be continuously displayed
%    Mproc - the number of threads to use
%    gamma - the highest level DP concentration parameter
%    alpha - the document level DP concentration parameter
%    endtime - the total time in seconds to stop after
%    numits - the number of iterations to stop after
%    isBars - visualize things specific to the bars dataset
%
%   Notes:
%     (1) The display shows the running time of the algorithm without I/O
%     to and from Matlab. It also doesn't show time to display. This is
%     more accurate, since one can always run as many iterations with C++
%     as desired.
%
%   [1] J. Chang and J. W. Fisher II, "Parallel Sampling of HDPs using
%       Sub-Cluster Splits". Neural Information Processing Systems (NIPS
%       2014), Montreal, Canada, Dec 2014.
%
%   Copyright(c) 2014. Jason Chang, CSAIL, MIT. 

if (~exist('isBars', 'var') || isempty(isBars))
    isBars = false;
end

addpath('include');

K = start;

for d=1:numel(data)
    N = numel(data(d).words);
    data(d).words = data(d).words(randperm(N));
end

params.its_crf = 20;
params.gamma = gamma; % hyper-parameter for highest DP
params.alpha = alpha; % hyper-parameter for document-level DP
params.diralpha = 0.5; % hyper-parameter for Dirichlet base measure
params.NThreads = Mproc;
params.HOW = 1;
params.W = dictSize;
params.always_splittable = logical(false);
params.Nsubs = 2;

load 'bestNumTables';
params.bestNumTables = bestNumTables;

maxW = max(squeeze(cellfun(@numel,struct2cell(data))));
params.logstirling = computeLogStirlingTable(maxW);

% Create the model struct.
model = struct('alpha',params.alpha,'gamma',params.gamma,'dishes',[], 'restaurants',[]);
model.dishes = struct('dishInd',[],'beta',[],'betaSubs',[], ...
    'subloglikelihoodOld',[],'subloglikelihoodDelta',[],'splittable',[]);
for k=1:K
    model.dishes(k).dishInd = int32(k-1);
    model.dishes(k).beta = 100/(100*(K+1));
    model.dishes(k).betaSubs = ones(params.Nsubs,1)/params.Nsubs;
    model.dishes(k).subloglikelihoodOld = -inf;
    model.dishes(k).subloglikelihoodDelta = inf;
    model.dishes(k).splittable = false;
    model.dishes(k).loglikelihoodOld = -inf;
    model.dishes(k).loglikelihoodDelta = inf;
    model.dishes(k).mergeable = false;
end

model.restaurants = struct('customers',[],'dishCounts',[]);
model.restaurants(numel(data)).customers = [];
for i=1:numel(data)
    dims = size(data(i).words);
    dims(2) = dims(2) - params.HOW;
    model.restaurants(i).customers = int32(randi(K,dims))-1;
    model.restaurants(i).customers = model.restaurants(i).customers*params.Nsubs + int32(randi(params.Nsubs,dims)-1);

    uvals = unique(floor(double(model.restaurants(i).customers)/2));
    dishCounts = ones(2,numel(uvals),'int32');
    dishCounts(1,:) = uvals;
    model.restaurants(i).dishCounts = dishCounts;
end

time = zeros(numits*params.its_crf+1,1);
E = zeros(numits*params.its_crf+1,1);
EHOW = zeros(numits*params.its_crf+1,1);
numTopics = zeros(numits*params.its_crf+1,1);
nhists = cell(numits*params.its_crf+1,1);

% Initialize the likelihood values.
cindex = 1;
[E(1), EHOW(1), numTopics(1), nhists(1)] = hdpmm_subclusters_calc_likelihood(data, params, model);

for it=1:numits
    [model, timediffs, Es, EHOWs, numTopicss, nhistss] = hdpmm_subclusters(data, params, model);
    time(cindex+1:cindex+params.its_crf) = time(cindex) + cumsum(timediffs);
    E(cindex+1:cindex+params.its_crf) = Es;
    EHOW(cindex+1:cindex+params.its_crf) = EHOWs;
    numTopics(cindex+1:cindex+params.its_crf) = numTopicss;
    nhists(cindex+1:cindex+params.its_crf) = nhistss;
    cindex = cindex+params.its_crf;
        
    if (time(cindex)>endtime)
        break;
    end
    
    % Display some progress stuff.
    if (dispOn)
        if (it == 1)
            figure(1);
            figure(2);
        end
        
        sfigure(1);
        subplot(3,1,1);
        plot(time(2:cindex),E(2:cindex));
        title('Joint Log Likelihood vs time');
        subplot(3,1,2);
        plot(time(1:cindex),EHOW(1:cindex));
        title('HOW Log Likelihood vs time');
        subplot(3,1,3);
        plot(time(1:cindex),numTopics(1:cindex));
        title('Number of Topics vs. time');
        drawnow;
        
        if (isBars)
            K = params.Nsubs*200;
            histograms = zeros(25,K);
            c = double([model.restaurants.customers]+1);
            w = zeros(size(c));
            ind = 1;
            for d=1:numel(data)
                Nd = numel(data(d).words) - params.HOW;
                newind = ind + Nd;
                w(ind:newind-1) = data(d).words(params.HOW+1:end)+1;
                ind = newind;
            end
            histograms = hist(sub2ind(size(histograms),w,c), 1:25*K);
            histograms = reshape(histograms,25,K);
            sfigure(2);
            pos = get(sfigure(2),'Position');
            pos(3) = 550;
            pos(4) = 550;
            set(sfigure(2),'Position', pos);
            clf
            count = 1;
            for k=1:params.Nsubs:size(histograms,2)
                if (any(any(histograms(:,k:k+params.Nsubs-1))>0))
                    subplot(7,2*(params.Nsubs+1)+1,count);
                    imagesc(reshape(sum(histograms(:,k:k+params.Nsubs-1),2),[5 5]));axis off;
                    
                    for s=0:params.Nsubs-1
                        count = count + 1;
                        subplot(7,2*(params.Nsubs+1)+1,count);
                        imagesc(reshape(histograms(:,k+s),[5 5]));axis off;
                    end
                    
                    if (mod(count,2*(params.Nsubs+1)+1)==0)
                        count = count + 1;
                    else
                        count = count + 2;
                    end
                    if (count>70)
                        break;
                    end
                end
            end
            subplot(7,2*(params.Nsubs+1)+1,1);
            title('Topics');
            subplot(7,2*(params.Nsubs+1)+1,2);
            title('Sub-Topic');
            subplot(7,2*(params.Nsubs+1)+1,3);
            title('Sub-Topic');
            
            subplot(7,2*(params.Nsubs+1)+1,5);
            title('Topics');
            subplot(7,2*(params.Nsubs+1)+1,6);
            title('Sub-Topic');
            subplot(7,2*(params.Nsubs+1)+1,7);
            title('Sub-Topic');
        end
        drawnow();
    end
    
    disp([num2str(cindex, '%04d') ' its - ' num2str(mean(time(2:cindex)-time(1:cindex-1)),'%0.4f') ' secs/it - Time: ' num2str(time(cindex)) ' secs']);
end

E = E(1:cindex);
EHOW = EHOW(1:cindex);
time = time(1:cindex);
numTopics = numTopics(1:cindex);
nhists = nhists(1:cindex);

